import time


class CloudInit:
    def __init__(self, container, config):
        self.container = container
        self.config = config
        self.CONFIGPATH = "/etc/cloud-init"

    def apply_config(self):
        self.cleanup(self.config.keys())

        for key, value in self.config.items():
            fpath = "%s/%s" % (self.CONFIGPATH, key)
            self.container.upload_content(fpath, value)
        if not self.is_running():
            self.start()

    def cleanup(self, macaddresses):
        configs = self.container.client.filesystem.list(self.CONFIGPATH)
        for config in configs:
            if config["name"] not in macaddresses:
                self.container.client.filesystem.remove("%s/%s" % (self.CONFIGPATH, config["name"]))

    def start(self):
        if not self.is_running():
            self.container.client.system(
                'cloud-init-server \
                -bind 127.0.0.1:8080 \
                -config {config}'
                .format(config=self.CONFIGPATH)
            )

        start = time.time()
        while time.time() + 10 > start:
            if self.is_running():
                return
            time.sleep(0.5)
        raise RuntimeError('Failed to start cloudinit server')

    def is_running(self):
        for port in self.container.client.info.port():
            if port['network'] == 'tcp' and port['port'] == 8080 and port['ip'] == '127.0.0.1':
                return True
        return False



from js9 import j


class Mountable:
    """
    Abstract implementation for devices that are mountable.
    Device should have attributes devicename and mountpoint
    """

    def mount(self, target, options=['defaults']):
        """
        @param target: Mount point
        @param options: Optional mount options
        """
        if self.mountpoint == target:
            return

        self._client.bash('mkdir -p {}'.format(target))

        self._client.disk.mount(
            source=self.devicename,
            target=target,
            options=options,
        )

        self.mountpoint = target

    def umount(self):
        """
        Unmount disk
        """
        if self.mountpoint:
            self._client.disk.umount(
                source=self.mountpoint,
            )
        self.mountpoint = None


class AYSable:
    """
    Abstract implementation for class that reflect an AYS service.
    class should have a name and actor attributes

    This provide common method to CRUD AYS service from the python classes
    """
    @property
    def name(self):
        return self._obj.name

    @property
    def role(self):
        return self.actor.split('.')[0]

    def create(self, aysrepo):
        """
        create the AYS Service
        """
        raise NotImplementedError()

    def get(self, aysrepo):
        """
        get the AYS service
        """
        try:
            return aysrepo.serviceGet(role=self.role, instance=self.name)
        except j.exceptions.NotFound:
            raise ValueError("Could not find {} with name {}".format(self.role, self.name))



"""
Auto-generated class for VMInfo
"""
from .VMDiskInfo import VMDiskInfo
from .VMNicInfo import VMNicInfo

from . import client_support


class VMInfo(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(cpu, disks, nics):
        """
        :type cpu: list[float]
        :type disks: list[VMDiskInfo]
        :type nics: list[VMNicInfo]
        :rtype: VMInfo
        """

        return VMInfo(
            cpu=cpu,
            disks=disks,
            nics=nics,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'VMInfo'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'cpu'
        val = data.get(property_name)
        if val is not None:
            datatypes = [float]
            try:
                self.cpu = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'disks'
        val = data.get(property_name)
        if val is not None:
            datatypes = [VMDiskInfo]
            try:
                self.disks = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'nics'
        val = data.get(property_name)
        if val is not None:
            datatypes = [VMNicInfo]
            try:
                self.nics = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



"""
Auto-generated class for Cluster
"""
from .EnumClusterDriveType import EnumClusterDriveType
from .EnumClusterStatus import EnumClusterStatus
from .StorageServer import StorageServer

from . import client_support


class Cluster(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(dataStorage, driveType, label, metadataStorage, nodes, status):
        """
        :type dataStorage: list[StorageServer]
        :type driveType: EnumClusterDriveType
        :type label: str
        :type metadataStorage: list[StorageServer]
        :type nodes: list[str]
        :type status: EnumClusterStatus
        :rtype: Cluster
        """

        return Cluster(
            dataStorage=dataStorage,
            driveType=driveType,
            label=label,
            metadataStorage=metadataStorage,
            nodes=nodes,
            status=status,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'Cluster'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'dataStorage'
        val = data.get(property_name)
        if val is not None:
            datatypes = [StorageServer]
            try:
                self.dataStorage = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'driveType'
        val = data.get(property_name)
        if val is not None:
            datatypes = [EnumClusterDriveType]
            try:
                self.driveType = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'label'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.label = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'metadataStorage'
        val = data.get(property_name)
        if val is not None:
            datatypes = [StorageServer]
            try:
                self.metadataStorage = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'nodes'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.nodes = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'status'
        val = data.get(property_name)
        if val is not None:
            datatypes = [EnumClusterStatus]
            try:
                self.status = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)



class NodesService:
    def __init__(self, client):
        self.client = client



    def DeleteBridge(self, bridgeid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Remove bridge
        It is method for DELETE /nodes/{nodeid}/bridges/{bridgeid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/bridges/"+bridgeid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetBridge(self, bridgeid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get bridge details
        It is method for GET /nodes/{nodeid}/bridges/{bridgeid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/bridges/"+bridgeid
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListBridges(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List bridges
        It is method for GET /nodes/{nodeid}/bridges
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/bridges"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateBridge(self, data, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Creates a new bridge
        It is method for POST /nodes/{nodeid}/bridges
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/bridges"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetContainerCPUInfo(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of all CPUs in the container
        It is method for GET /nodes/{nodeid}/containers/{containername}/cpus
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/cpus"
        return self.client.get(uri, None, headers, query_params, content_type)


    def GetContainerDiskInfo(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of all the disks in the container
        It is method for GET /nodes/{nodeid}/containers/{containername}/disks
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/disks"
        return self.client.get(uri, None, headers, query_params, content_type)


    def FileDelete(self, data, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete file from container
        It is method for DELETE /nodes/{nodeid}/containers/{containername}/filesystem
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/filesystem"
        return self.client.delete(uri, data, headers, query_params, content_type)


    def FileDownload(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Download file from container
        It is method for GET /nodes/{nodeid}/containers/{containername}/filesystem
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/filesystem"
        return self.client.get(uri, None, headers, query_params, content_type)


    def FileUpload(self, data, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Upload file to container
        It is method for POST /nodes/{nodeid}/containers/{containername}/filesystem
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/filesystem"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetContainerOSInfo(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of the container OS
        It is method for GET /nodes/{nodeid}/containers/{containername}/info
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/info"
        return self.client.get(uri, None, headers, query_params, content_type)


    def KillContainerJob(self, jobid, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Kills the job
        It is method for DELETE /nodes/{nodeid}/containers/{containername}/jobs/{jobid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/jobs/"+jobid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetContainerJob(self, jobid, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get details of a submitted job on the container
        It is method for GET /nodes/{nodeid}/containers/{containername}/jobs/{jobid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/jobs/"+jobid
        return self.client.get(uri, None, headers, query_params, content_type)


    def SendSignalToJob(self, data, jobid, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Send signal to the job
        It is method for POST /nodes/{nodeid}/containers/{containername}/jobs/{jobid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/jobs/"+jobid
        return self.client.post(uri, data, headers, query_params, content_type)


    def KillAllContainerJobs(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Kill all running jobs on the container
        It is method for DELETE /nodes/{nodeid}/containers/{containername}/jobs
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/jobs"
        return self.client.delete(uri, None, headers, query_params, content_type)


    def ListContainerJobs(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List running jobs on the container
        It is method for GET /nodes/{nodeid}/containers/{containername}/jobs
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/jobs"
        return self.client.get(uri, None, headers, query_params, content_type)


    def StartContainerJob(self, data, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Start a new job in this container
        It is method for POST /nodes/{nodeid}/containers/{containername}/jobs
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/jobs"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetContainerMemInfo(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information about the memory in the container
        It is method for GET /nodes/{nodeid}/containers/{containername}/mem
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/mem"
        return self.client.get(uri, None, headers, query_params, content_type)


    def GetContainerNicInfo(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information about the network interfaces in the container
        It is method for GET /nodes/{nodeid}/containers/{containername}/nics
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/nics"
        return self.client.get(uri, None, headers, query_params, content_type)


    def PingContainer(self, data, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Ping this container
        It is method for POST /nodes/{nodeid}/containers/{containername}/ping
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/ping"
        return self.client.post(uri, data, headers, query_params, content_type)


    def KillContainerProcess(self, processid, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Kills the process by sending sigterm signal to the process. If it is still running, a sigkill signal will be sent to the process
        It is method for DELETE /nodes/{nodeid}/containers/{containername}/processes/{processid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/processes/"+processid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetContainerProcess(self, processid, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get process details
        It is method for GET /nodes/{nodeid}/containers/{containername}/processes/{processid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/processes/"+processid
        return self.client.get(uri, None, headers, query_params, content_type)


    def SendSignalToProcess(self, data, processid, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Send signal to the process
        It is method for POST /nodes/{nodeid}/containers/{containername}/processes/{processid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/processes/"+processid
        return self.client.post(uri, data, headers, query_params, content_type)


    def ListContainerProcesses(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get running processes in this container
        It is method for GET /nodes/{nodeid}/containers/{containername}/processes
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/processes"
        return self.client.get(uri, None, headers, query_params, content_type)


    def StartContainer(self, data, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Start container instance
        It is method for POST /nodes/{nodeid}/containers/{containername}/start
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/start"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetContainerState(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get aggregated consumption of container + all processes (CPU, memory, etc.)
        It is method for GET /nodes/{nodeid}/containers/{containername}/state
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/state"
        return self.client.get(uri, None, headers, query_params, content_type)


    def StopContainer(self, data, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Stop container instance
        It is method for POST /nodes/{nodeid}/containers/{containername}/stop
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername+"/stop"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteContainer(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete container instance
        It is method for DELETE /nodes/{nodeid}/containers/{containername}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetContainer(self, containername, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get container
        It is method for GET /nodes/{nodeid}/containers/{containername}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers/"+containername
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListContainers(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List running containers
        It is method for GET /nodes/{nodeid}/containers
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateContainer(self, data, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Create a new container
        It is method for POST /nodes/{nodeid}/containers
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/containers"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetCPUInfo(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of all CPUs in the node
        It is method for GET /nodes/{nodeid}/cpus
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/cpus"
        return self.client.get(uri, None, headers, query_params, content_type)


    def GetDiskInfo(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of all the disks in the node
        It is method for GET /nodes/{nodeid}/disks
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/disks"
        return self.client.get(uri, None, headers, query_params, content_type)


    def GetGWFWConfig(self, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get current FW config
        It is method for GET /nodes/{nodeid}/gws/{gwname}/advanced/firewall
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/advanced/firewall"
        return self.client.get(uri, None, headers, query_params, content_type)


    def SetGWFWConfig(self, data, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Set FW config
        Once used you can not use gw.portforwards any longer
        It is method for POST /nodes/{nodeid}/gws/{gwname}/advanced/firewall
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/advanced/firewall"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetGWHTTPConfig(self, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get current HTTP config
        It is method for GET /nodes/{nodeid}/gws/{gwname}/advanced/http
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/advanced/http"
        return self.client.get(uri, None, headers, query_params, content_type)


    def SetGWHTTPConfig(self, data, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Set HTTP config
        Once used you can not use gw.httpproxxies any longer
        It is method for POST /nodes/{nodeid}/gws/{gwname}/advanced/http
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/advanced/http"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteDHCPHost(self, macaddress, interface, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete dhcp host
        It is method for DELETE /nodes/{nodeid}/gws/{gwname}/dhcp/{interface}/hosts/{macaddress}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/dhcp/"+interface+"/hosts/"+macaddress
        return self.client.delete(uri, None, headers, query_params, content_type)


    def ListGWDHCPHosts(self, interface, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List DHCPHosts for specified interface
        It is method for GET /nodes/{nodeid}/gws/{gwname}/dhcp/{interface}/hosts
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/dhcp/"+interface+"/hosts"
        return self.client.get(uri, None, headers, query_params, content_type)


    def AddGWDHCPHost(self, data, interface, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Add a dhcp host to a specified interface
        It is method for POST /nodes/{nodeid}/gws/{gwname}/dhcp/{interface}/hosts
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/dhcp/"+interface+"/hosts"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteGWForward(self, forwardid, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete portforward, forwardid = srcip:srcport
        It is method for DELETE /nodes/{nodeid}/gws/{gwname}/firewall/forwards/{forwardid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/firewall/forwards/"+forwardid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetGWForwards(self, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get list for IPv4 Forwards
        It is method for GET /nodes/{nodeid}/gws/{gwname}/firewall/forwards
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/firewall/forwards"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateGWForwards(self, data, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Create a new Portforwarding
        It is method for POST /nodes/{nodeid}/gws/{gwname}/firewall/forwards
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/firewall/forwards"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteHTTPProxies(self, proxyid, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete HTTP proxy
        It is method for DELETE /nodes/{nodeid}/gws/{gwname}/httpproxies/{proxyid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/httpproxies/"+proxyid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetHTTPProxy(self, proxyid, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get info of HTTP proxy
        It is method for GET /nodes/{nodeid}/gws/{gwname}/httpproxies/{proxyid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/httpproxies/"+proxyid
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListHTTPProxies(self, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List for HTTP proxies
        It is method for GET /nodes/{nodeid}/gws/{gwname}/httpproxies
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/httpproxies"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateHTTPProxies(self, data, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Create new HTTP proxies
        It is method for POST /nodes/{nodeid}/gws/{gwname}/httpproxies
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/httpproxies"
        return self.client.post(uri, data, headers, query_params, content_type)


    def StartGateway(self, data, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Start Gateway instance
        It is method for POST /nodes/{nodeid}/gws/{gwname}/start
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/start"
        return self.client.post(uri, data, headers, query_params, content_type)


    def StopGateway(self, data, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Stop gateway instance
        It is method for POST /nodes/{nodeid}/gws/{gwname}/stop
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname+"/stop"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteGateway(self, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete gateway instance
        It is method for DELETE /nodes/{nodeid}/gws/{gwname}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetGateway(self, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get gateway
        It is method for GET /nodes/{nodeid}/gws/{gwname}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname
        return self.client.get(uri, None, headers, query_params, content_type)


    def UpdateGateway(self, data, gwname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Update Gateway
        It is method for PUT /nodes/{nodeid}/gws/{gwname}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws/"+gwname
        return self.client.put(uri, data, headers, query_params, content_type)


    def ListGateways(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List running gateways
        It is method for GET /nodes/{nodeid}/gws
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateGW(self, data, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Create a new gateway
        It is method for POST /nodes/{nodeid}/gws
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/gws"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetNodeOSInfo(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of the OS of the node
        It is method for GET /nodes/{nodeid}/info
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/info"
        return self.client.get(uri, None, headers, query_params, content_type)


    def KillNodeJob(self, jobid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Kills the job
        It is method for DELETE /nodes/{nodeid}/jobs/{jobid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/jobs/"+jobid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetNodeJob(self, jobid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get the details of a submitted job
        It is method for GET /nodes/{nodeid}/jobs/{jobid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/jobs/"+jobid
        return self.client.get(uri, None, headers, query_params, content_type)


    def KillAllNodeJobs(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Kill all running jobs
        It is method for DELETE /nodes/{nodeid}/jobs
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/jobs"
        return self.client.delete(uri, None, headers, query_params, content_type)


    def ListNodeJobs(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List running jobs
        It is method for GET /nodes/{nodeid}/jobs
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/jobs"
        return self.client.get(uri, None, headers, query_params, content_type)


    def GetMemInfo(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information about the memory in the node
        It is method for GET /nodes/{nodeid}/mem
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/mem"
        return self.client.get(uri, None, headers, query_params, content_type)


    def GetNodeMounts(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of all the mountpoints on the node
        It is method for GET /nodes/{nodeid}/mounts
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/mounts"
        return self.client.get(uri, None, headers, query_params, content_type)


    def GetNicInfo(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information about the network interfaces in the node
        It is method for GET /nodes/{nodeid}/nics
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/nics"
        return self.client.get(uri, None, headers, query_params, content_type)


    def PingNode(self, data, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Ping this node
        It is method for POST /nodes/{nodeid}/ping
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/ping"
        return self.client.post(uri, data, headers, query_params, content_type)


    def KillNodeProcess(self, processid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Kills the process by sending sigterm signal to the process. If it is still running, a sigkill signal will be sent to the process
        It is method for DELETE /nodes/{nodeid}/processes/{processid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/processes/"+processid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetNodeProcess(self, processid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get process details
        It is method for GET /nodes/{nodeid}/processes/{processid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/processes/"+processid
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListNodeProcesses(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get processes
        It is method for GET /nodes/{nodeid}/processes
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/processes"
        return self.client.get(uri, None, headers, query_params, content_type)


    def RebootNode(self, data, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Immediately reboot the machine
        It is method for POST /nodes/{nodeid}/reboot
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/reboot"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetNodeState(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        The aggregated consumption of node + all processes (cpu, memory, etc...)
        It is method for GET /nodes/{nodeid}/state
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/state"
        return self.client.get(uri, None, headers, query_params, content_type)


    def DeleteStoragePoolDevice(self, deviceuuid, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Removes the device from the storage pool
        It is method for DELETE /nodes/{nodeid}/storagepools/{storagepoolname}/devices/{deviceuuid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/devices/"+deviceuuid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetStoragePoolDeviceInfo(self, deviceuuid, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get information of the device
        It is method for GET /nodes/{nodeid}/storagepools/{storagepoolname}/devices/{deviceuuid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/devices/"+deviceuuid
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListStoragePoolDevices(self, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List the devices in the storage pool
        It is method for GET /nodes/{nodeid}/storagepools/{storagepoolname}/devices
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/devices"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateStoragePoolDevices(self, data, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Add extra devices to this storage pool
        It is method for POST /nodes/{nodeid}/storagepools/{storagepoolname}/devices
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/devices"
        return self.client.post(uri, data, headers, query_params, content_type)


    def RollbackFilesystemSnapshot(self, data, snapshotname, filesystemname, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Rollback the file system to the state at the moment the snapshot was taken
        It is method for POST /nodes/{nodeid}/storagepools/{storagepoolname}/filesystems/{filesystemname}/snapshots/{snapshotname}/rollback
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/filesystems/"+filesystemname+"/snapshots/"+snapshotname+"/rollback"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteFilesystemSnapshot(self, snapshotname, filesystemname, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete snapshot
        It is method for DELETE /nodes/{nodeid}/storagepools/{storagepoolname}/filesystems/{filesystemname}/snapshots/{snapshotname}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/filesystems/"+filesystemname+"/snapshots/"+snapshotname
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetFilesystemSnapshotInfo(self, snapshotname, filesystemname, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information on the snapshot
        It is method for GET /nodes/{nodeid}/storagepools/{storagepoolname}/filesystems/{filesystemname}/snapshots/{snapshotname}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/filesystems/"+filesystemname+"/snapshots/"+snapshotname
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListFilesystemSnapshots(self, filesystemname, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List snapshots of this file system
        It is method for GET /nodes/{nodeid}/storagepools/{storagepoolname}/filesystems/{filesystemname}/snapshots
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/filesystems/"+filesystemname+"/snapshots"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateSnapshot(self, data, filesystemname, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Create a new read-only snapshot of the current state of the vdisk
        It is method for POST /nodes/{nodeid}/storagepools/{storagepoolname}/filesystems/{filesystemname}/snapshots
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/filesystems/"+filesystemname+"/snapshots"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteFilesystem(self, filesystemname, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete file system
        It is method for DELETE /nodes/{nodeid}/storagepools/{storagepoolname}/filesystems/{filesystemname}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/filesystems/"+filesystemname
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetFilesystemInfo(self, filesystemname, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed file system information
        It is method for GET /nodes/{nodeid}/storagepools/{storagepoolname}/filesystems/{filesystemname}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/filesystems/"+filesystemname
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListFilesystems(self, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List all file systems
        It is method for GET /nodes/{nodeid}/storagepools/{storagepoolname}/filesystems
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/filesystems"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateFilesystem(self, data, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Create a new file system
        It is method for POST /nodes/{nodeid}/storagepools/{storagepoolname}/filesystems
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname+"/filesystems"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteStoragePool(self, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete the storage pool
        It is method for DELETE /nodes/{nodeid}/storagepools/{storagepoolname}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetStoragePoolInfo(self, storagepoolname, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of this storage pool
        It is method for GET /nodes/{nodeid}/storagepools/{storagepoolname}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools/"+storagepoolname
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListStoragePools(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List storage pools present in the node
        It is method for GET /nodes/{nodeid}/storagepools
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateStoragePool(self, data, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Create a new storage pool in the node
        It is method for POST /nodes/{nodeid}/storagepools
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/storagepools"
        return self.client.post(uri, data, headers, query_params, content_type)


    def GetVMInfo(self, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get statistical information about the virtual machine.
        It is method for GET /nodes/{nodeid}/vms/{vmid}/info
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid+"/info"
        return self.client.get(uri, None, headers, query_params, content_type)


    def MigrateVM(self, data, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Migrate the virtual machine to another host
        It is method for POST /nodes/{nodeid}/vms/{vmid}/migrate
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid+"/migrate"
        return self.client.post(uri, data, headers, query_params, content_type)


    def PauseVM(self, data, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Pauses the VM
        It is method for POST /nodes/{nodeid}/vms/{vmid}/pause
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid+"/pause"
        return self.client.post(uri, data, headers, query_params, content_type)


    def ResumeVM(self, data, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Resumes the virtual machine
        It is method for POST /nodes/{nodeid}/vms/{vmid}/resume
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid+"/resume"
        return self.client.post(uri, data, headers, query_params, content_type)


    def ShutdownVM(self, data, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Gracefully shutdown the virtual machine
        It is method for POST /nodes/{nodeid}/vms/{vmid}/shutdown
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid+"/shutdown"
        return self.client.post(uri, data, headers, query_params, content_type)


    def StartVM(self, data, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Start the virtual machine
        It is method for POST /nodes/{nodeid}/vms/{vmid}/start
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid+"/start"
        return self.client.post(uri, data, headers, query_params, content_type)


    def StopVM(self, data, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Stops the VM
        It is method for POST /nodes/{nodeid}/vms/{vmid}/stop
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid+"/stop"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteVM(self, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Deletes the virtual machine
        It is method for DELETE /nodes/{nodeid}/vms/{vmid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetVM(self, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get the virtual machine object
        It is method for GET /nodes/{nodeid}/vms/{vmid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid
        return self.client.get(uri, None, headers, query_params, content_type)


    def UpdateVM(self, data, vmid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Updates the virtual machine
        It is method for PUT /nodes/{nodeid}/vms/{vmid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms/"+vmid
        return self.client.put(uri, data, headers, query_params, content_type)


    def ListVMs(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List all virtual machines
        It is method for GET /nodes/{nodeid}/vms
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateVM(self, data, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Creates a new virtual machine
        It is method for POST /nodes/{nodeid}/vms
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/vms"
        return self.client.post(uri, data, headers, query_params, content_type)


    def ExitZerotier(self, zerotierid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Exit the ZeroTier network
        It is method for DELETE /nodes/{nodeid}/zerotiers/{zerotierid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/zerotiers/"+zerotierid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetZerotier(self, zerotierid, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get ZeroTier network details
        It is method for GET /nodes/{nodeid}/zerotiers/{zerotierid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/zerotiers/"+zerotierid
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListZerotier(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        List running ZeroTier networks
        It is method for GET /nodes/{nodeid}/zerotiers
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/zerotiers"
        return self.client.get(uri, None, headers, query_params, content_type)


    def JoinZerotier(self, data, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Join ZeroTier network
        It is method for POST /nodes/{nodeid}/zerotiers
        """
        uri = self.client.base_url + "/nodes/"+nodeid+"/zerotiers"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteNode(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete a node
        It is method for DELETE /nodes/{nodeid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetNode(self, nodeid, headers=None, query_params=None, content_type="application/json"):
        """
        Get detailed information of a node
        It is method for GET /nodes/{nodeid}
        """
        uri = self.client.base_url + "/nodes/"+nodeid
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListNodes(self, headers=None, query_params=None, content_type="application/json"):
        """
        List all nodes
        It is method for GET /nodes
        """
        uri = self.client.base_url + "/nodes"
        return self.client.get(uri, None, headers, query_params, content_type)



class VdisksService:
    def __init__(self, client):
        self.client = client



    def ResizeVdisk(self, data, vdiskid, headers=None, query_params=None, content_type="application/json"):
        """
        Resize vdisk
        It is method for POST /vdisks/{vdiskid}/resize
        """
        uri = self.client.base_url + "/vdisks/"+vdiskid+"/resize"
        return self.client.post(uri, data, headers, query_params, content_type)


    def RollbackVdisk(self, data, vdiskid, headers=None, query_params=None, content_type="application/json"):
        """
        Rollback a vdisk to a previous state
        It is method for POST /vdisks/{vdiskid}/rollback
        """
        uri = self.client.base_url + "/vdisks/"+vdiskid+"/rollback"
        return self.client.post(uri, data, headers, query_params, content_type)


    def DeleteVdisk(self, vdiskid, headers=None, query_params=None, content_type="application/json"):
        """
        Delete Vdisk
        It is method for DELETE /vdisks/{vdiskid}
        """
        uri = self.client.base_url + "/vdisks/"+vdiskid
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetVdiskInfo(self, vdiskid, headers=None, query_params=None, content_type="application/json"):
        """
        Get vdisk information
        It is method for GET /vdisks/{vdiskid}
        """
        uri = self.client.base_url + "/vdisks/"+vdiskid
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListVdisks(self, headers=None, query_params=None, content_type="application/json"):
        """
        List vdisks
        It is method for GET /vdisks
        """
        uri = self.client.base_url + "/vdisks"
        return self.client.get(uri, None, headers, query_params, content_type)


    def CreateNewVdisk(self, data, headers=None, query_params=None, content_type="application/json"):
        """
        Create a new vdisk, can be a copy from an existing vdisk
        It is method for POST /vdisks
        """
        uri = self.client.base_url + "/vdisks"
        return self.client.post(uri, data, headers, query_params, content_type)



from .abstracts import Mountable
import os
import time

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def _prepare_device(node, devicename):
    logger.debug("prepare device %s", devicename)
    ss = devicename.split('/')
    if len(ss) < 3:
        raise RuntimeError("bad device name: {}".format(devicename))
    name = ss[2]

    disk = node.disks.get(name)
    if disk is None:
        raise ValueError("device {} not found".format(name))

    node.client.system('parted -s /dev/{} mklabel gpt mkpart primary 1m 100%'.format(name)).get()
    now = time.time()
    # check partitions is ready and writable
    while now + 60 > time.time():
        try:
            disk = node.disks.get(name)
            if len(disk.partitions) > 0:
                partition = disk.partitions[0]
                resp = node.client.bash('test -b {0} && dd if={0} of=/dev/null bs=4k count=1024'.format(partition.devicename)).get()
                if resp.state == 'SUCCESS':
                    return partition
        except:
            time.sleep(1)
            continue
    else:
        raise RuntimeError("Failed to create partition")


class StoragePools:
    def __init__(self, node):
        self.node = node
        self._client = node._client

    def list(self):
        storagepools = []
        btrfs_list = self._client.btrfs.list()
        if btrfs_list:
            for btrfs in self._client.btrfs.list():
                if btrfs['label'].startswith('sp_'):
                    name = btrfs['label'].split('_', 1)[1]
                    devicenames = [device['path'] for device in btrfs['devices']]
                    storagepools.append(StoragePool(self.node, name, devicenames))
        return storagepools

    def get(self, name):
        for pool in self.list():
            if pool.name == name:
                return pool
        raise ValueError("Could not find StoragePool with name {}".format(name))

    def create(self, name, devices, metadata_profile, data_profile, overwrite=False):
        label = 'sp_{}'.format(name)
        logger.debug("create storagepool %s", label)

        device_names = []
        for device in devices:
            part = _prepare_device(self.node, device)
            device_names.append(part.devicename)

        self._client.btrfs.create(label, device_names, metadata_profile, data_profile, overwrite=overwrite)
        pool = StoragePool(self.node, name, device_names)
        return pool


class StoragePool(Mountable):

    def __init__(self, node, name, devices):
        self.node = node
        self._client = node._client
        self.devices = devices
        self.name = name
        self._mountpoint = None
        self._ays = None

    @property
    def devicename(self):
        return 'UUID={}'.format(self.uuid)

    def mount(self, target=None):
        if target is None:
            target = os.path.join('/mnt/storagepools/{}'.format(self.name))
        return super().mount(target)

    def delete(self, zero=True):
        """
        Destroy storage pool
        param name: name of storagepool to delete
        param zero: write zeros (nulls) to the first 500MB of each disk in this storagepool
        """

        if self.mountpoint:
            self.umount()
        if zero:
            partitionmap = {}
            for disk in self.node.disks.list():
                for partition in disk.partitions:
                    partitionmap[partition.name] = partition
            for device in self.devices:
                diskpath = os.path.basename(device)
                partition = partitionmap.get(diskpath)
                if partition:
                    disk = partition.disk
                    self._client.disk.rmpart(disk.name, 1)
                    self._client.system('dd if=/dev/zero bs=1M count=500 of={}'.format(disk.devicename)).get()

    @property
    def mountpoint(self):
        mounts = self.node.list_mounts()
        for device in self.devices:
            for mount in mounts:
                if mount.device == device:
                    options = mount.options.split(',')
                    if 'subvol=/' in options:
                        return mount.mountpoint

    def is_device_used(self, device):
        """
        check if the device passed as argument is already part of this storagepool
        @param device: str e.g: /dev/sda
        """
        for d in self.devices:
            if d.startswith(device):
                return True
        return False

    def device_add(self, *devices):
        to_add = []
        for device in devices:
            if self.is_device_used(device):
                continue
            part = _prepare_device(self.node, device)
            logger.debug("add device %s to %s", device, self)
            to_add.append(part.devicename)

        self._client.btrfs.device_add(self._get_mountpoint(), *to_add)
        self.devices.extend(to_add)

    def device_remove(self, *devices):
        self._client.btrfs.device_remove(self._get_mountpoint(), *devices)
        for device in devices:
            if device in self.devices:
                logger.debug("remove device %s to %s", device, self)
                self.devices.remove(device)

    @property
    def fsinfo(self):
        if self.mountpoint is None:
            raise ValueError("can't get fsinfo if storagepool is not mounted")
        return self._client.btrfs.info(self.mountpoint)

    @mountpoint.setter
    def mountpoint(self, value):
        # do not do anything mountpoint is dynamic
        return

    def _get_mountpoint(self):
        mountpoint = self.mountpoint
        if not mountpoint:
            raise RuntimeError("Can not perform action when filesystem is not mounted")
        return mountpoint

    @property
    def info(self):
        for fs in self._client.btrfs.list():
            if fs['label'] == 'sp_{}'.format(self.name):
                return fs
        return None

    def raw_list(self):
        mountpoint = self._get_mountpoint()
        return self._client.btrfs.subvol_list(mountpoint) or []

    def get_devices_and_status(self):
        device_map = []
        disks = self._client.disk.list()['blockdevices']
        pool_status = 'healthy'
        for device in self.devices:
            info = None
            for disk in disks:
                disk_name = "/dev/%s" % disk['kname']
                if device == disk_name and disk['mountpoint']:
                    info = disk
                    break
                for part in disk.get('children', []) or []:
                    if device == "/dev/%s" % part['kname']:
                        info = part
                        break
                if info:
                    break

            status = 'healthy'
            if info['subsystems'] != 'block:virtio:pci':
                result = self._client.bash("smartctl -H %s > /dev/null ;echo $?" % disk_name).get()
                exit_status = int(result.stdout)

                if exit_status & 1 << 0:
                    status = "unknown"
                    pool_status = 'degraded'
                if (exit_status & 1 << 2) or (exit_status & 1 << 3):
                    status = 'degraded'
                    pool_status = 'degraded'

            device_map.append({
                'device': device,
                'partUUID': info['partuuid'] or '' if info else '',
                'status': status,
            })

        return device_map, pool_status

    def list(self):
        subvolumes = []
        for subvol in self.raw_list():
            path = subvol['Path']
            type_, _, name = path.partition('/')
            if type_ == 'filesystems':
                subvolumes.append(FileSystem(name, self))
        return subvolumes

    def get(self, name):
        """
        Get Filesystem
        """
        for filesystem in self.list():
            if filesystem.name == name:
                return filesystem
        raise ValueError("Could not find filesystem with name {}".format(name))

    def exists(self, name):
        """
        Check if filesystem with name exists
        """
        for subvolume in self.list():
            if subvolume.name == name:
                return True
        return False

    def create(self, name, quota=None):
        """
        Create filesystem
        """
        logger.debug("Create filesystem %s on %s", name, self)
        mountpoint = self._get_mountpoint()
        fspath = os.path.join(mountpoint, 'filesystems')
        self._client.filesystem.mkdir(fspath)
        subvolpath = os.path.join(fspath, name)
        self._client.btrfs.subvol_create(subvolpath)
        if quota:
            pass
        return FileSystem(name, self)

    @property
    def size(self):
        total = 0
        fs = self.info
        if fs:
            for device in fs['devices']:
                total += device['size']
        return total

    @property
    def uuid(self):
        fs = self.info
        if fs:
            return fs['uuid']
        return None

    @property
    def used(self):
        total = 0
        fs = self.info
        if fs:
            for device in fs['devices']:
                total += device['used']
        return total

    @property
    def ays(self):
        if self._ays is None:
            from zeroos.orchestrator.sal.atyourservice.StoragePool import StoragePoolAys
            self._ays = StoragePoolAys(self)
        return self._ays

    def __repr__(self):
        return "StoragePool <{}>".format(self.name)


class FileSystem:
    def __init__(self, name, pool):
        self.name = name
        self.pool = pool
        self._client = pool.node.client
        self.subvolume = "filesystems/{}".format(name)
        self.path = os.path.join(self.pool.mountpoint, self.subvolume)
        self.snapshotspath = os.path.join(self.pool.mountpoint, 'snapshots', self.name)
        self._ays = None

    def delete(self, includesnapshots=True):
        """
        Delete filesystem
        """
        self._client.btrfs.subvol_delete(self.path)
        if includesnapshots:
            for snapshot in self.list():
                snapshot.delete()
            self._client.filesystem.remove(self.snapshotspath)

    def get(self, name):
        """
        Get snapshot
        """
        for snap in self.list():
            if snap.name == name:
                return snap
        raise ValueError("Could not find snapshot {}".format(name))

    def list(self):
        """
        List snapshots
        """
        snapshots = []
        if self._client.filesystem.exists(self.snapshotspath):
            for fileentry in self._client.filesystem.list(self.snapshotspath):
                if fileentry['is_dir']:
                    snapshots.append(Snapshot(fileentry['name'], self))
        return snapshots

    def exists(self, name):
        """
        Check if a snapshot exists
        """
        return name in self.list()

    def create(self, name):
        """
        Create snapshot
        """
        logger.debug("create snapshot %s on %s", name, self.pool)
        snapshot = Snapshot(name, self)
        if self.exists(name):
            raise RuntimeError("Snapshot path {} exists.")
        self._client.filesystem.mkdir(self.snapshotspath)
        self._client.btrfs.subvol_snapshot(self.path, snapshot.path)
        return snapshot

    @property
    def ays(self):
        if self._ays is None:
            from JumpScale.sal.g8os.atyourservice.StoragePool import FileSystemAys
            self._ays = FileSystemAys(self)
        return self._ays

    def __repr__(self):
        return "FileSystem <{}: {!r}>".format(self.name, self.pool)


class Snapshot:
    def __init__(self, name, filesystem):
        self.filesystem = filesystem
        self._client = filesystem.pool.node.client
        self.name = name
        self.path = os.path.join(self.filesystem.snapshotspath, name)
        self.subvolume = "snapshots/{}/{}".format(self.filesystem.name, name)

    def rollback(self):
        self.filesystem.delete(False)
        self._client.btrfs.subvol_snapshot(self.path, self.filesystem.path)

    def delete(self):
        self._client.btrfs.subvol_delete(self.path)

    def __repr__(self):
        return "Snapshot <{}: {!r}>".format(self.name, self.filesystem)



from js9 import j
from .ARDB import ARDB

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class StorageCluster:
    """StorageCluster is a cluster of ardb servers"""

    def __init__(self, label, nodes=None, disk_type=None):
        """
        @param label: string repsenting the name of the storage cluster
        """
        self.label = label
        self.name = label
        self.nodes = nodes or []
        self.filesystems = []
        self.storage_servers = []
        self.disk_type = disk_type
        self.k = 0
        self.m = 0
        self._ays = None

    @classmethod
    def from_ays(cls, service, password):
        logger.debug("load cluster storage cluster from service (%s)", service)
        disk_type = str(service.model.data.diskType)

        nodes = []
        storage_servers = []
        for ardb_service in service.producers.get('ardb', []):
            storages_server = StorageServer.from_ays(ardb_service, password)
            storage_servers.append(storages_server)
            if storages_server.node not in nodes:
                nodes.append(storages_server.node)

        cluster = cls(label=service.name, nodes=nodes, disk_type=disk_type)
        cluster.storage_servers = storage_servers
        cluster.k = service.model.data.k
        cluster.m = service.model.data.m
        return cluster

    def get_config(self):
        data = {'dataStorage': [],
                'metadataStorage': None,
                'label': self.name,
                'status': 'ready' if self.is_running() else 'error',
                'nodes': [node.name for node in self.nodes]}
        for storageserver in self.storage_servers:
            if 'metadata' in storageserver.name:
                data['metadataStorage'] = {'address': storageserver.ardb.bind}
            else:
                data['dataStorage'].append({'address': storageserver.ardb.bind})
        return data

    @property
    def nr_server(self):
        """
        Number of storage server part of this cluster
        """
        return len(self.storage_servers)

    def find_disks(self):
        """
        return a list of disk that are not used by storage pool
        or has a different type as the one required for this cluster
        """
        logger.debug("find available_disks")
        cluster_name = 'sp_cluster_{}'.format(self.label)
        available_disks = {}

        def check_partition(disk):
            for partition in disk.partitions:
                for filesystem in partition.filesystems:
                    if filesystem['label'].startswith(cluster_name):
                        return True

        for node in self.nodes:
            for disk in node.disks.list():
                # skip disks of wrong type
                if disk.type.name != self.disk_type:
                    continue
                # skip devices which have filesystems on the device
                if len(disk.filesystems) > 0:
                    continue

                # include devices which have partitions
                if len(disk.partitions) == 0:
                    available_disks.setdefault(node.name, []).append(disk)
                else:
                    if check_partition(disk):
                        # devices that have partitions with correct label will be in the beginning
                        available_disks.setdefault(node.name, []).insert(0, disk)
        return available_disks

    def start(self):
        logger.debug("start %s", self)
        for server in self.storage_servers:
            server.start()

    def stop(self):
        logger.debug("stop %s", self)
        for server in self.storage_servers:
            server.stop()

    def is_running(self):
        # TODO: Improve this, what about part of server running and part stopped
        for server in self.storage_servers:
            if not server.is_running():
                return False
        return True

    def health(self):
        """
        Return a view of the state all storage server running in this cluster
        example :
        {
        'cluster1_1': {'ardb': True, 'container': True},
        'cluster1_2': {'ardb': True, 'container': True},
        }
        """
        health = {}
        for server in self.storage_servers:
            running, _ = server.ardb.is_running()
            health[server.name] = {
                'ardb': running,
                'container': server.container.is_running(),
            }
        return health

    def __str__(self):
        return "StorageCluster <{}>".format(self.label)

    def __repr__(self):
        return str(self)


class StorageServer:
    """ardb servers"""

    def __init__(self, cluster):
        self.cluster = cluster
        self.container = None
        self.ardb = None

    @classmethod
    def from_ays(cls, ardb_services, password=None):
        ardb = ARDB.from_ays(ardb_services, password)
        storage_server = cls(None)
        storage_server.container = ardb.container
        storage_server.ardb = ardb
        return storage_server

    @property
    def name(self):
        if self.ardb:
            return self.ardb.name
        return None

    @property
    def node(self):
        if self.container:
            return self.container.node
        return None

    def _find_port(self, start_port=2000):
        while True:
            if j.sal.nettools.tcpPortConnectionTest(self.node.addr, start_port, timeout=2):
                start_port += 1
                continue
            return start_port

    def start(self, timeout=30):
        logger.debug("start %s", self)
        if not self.container.is_running():
            self.container.start()

        ip, port = self.ardb.bind.split(":")
        self.ardb.bind = '{}:{}'.format(ip, self._find_port(port))
        self.ardb.start(timeout=timeout)

    def stop(self, timeout=30):
        logger.debug("stop %s", self)
        self.ardb.stop(timeout=timeout)
        self.container.stop()

    def is_running(self):
        container = self.container.is_running()
        ardb, _ = self.ardb.is_running()
        return (container and ardb)

    def __str__(self):
        return "StorageServer <{}>".format(self.container.name)

    def __repr__(self):
        return str(self)



class RunsService:
    def __init__(self, client):
        self.client = client



    def WaitOnRun(self, runid, headers=None, query_params=None, content_type="application/json"):
        """
        Wait for Run
        It is method for GET /runs/{runid}/wait
        """
        uri = self.client.base_url + "/runs/"+runid+"/wait"
        return self.client.get(uri, None, headers, query_params, content_type)


    def GetRunState(self, runid, headers=None, query_params=None, content_type="application/json"):
        """
        Get Run Status
        It is method for GET /runs/{runid}
        """
        uri = self.client.base_url + "/runs/"+runid
        return self.client.get(uri, None, headers, query_params, content_type)



from . import templates
import ipaddress


class Network:
    def __init__(self, iface, cidr):
        self.iface = iface
        ipiface = ipaddress.IPv4Interface(cidr)
        self.ipaddress = str(ipiface.ip)
        self.subnet = str(ipiface.network)


class Firewall:
    def __init__(self, container, publicnetwork, privatenetworks, forwards):
        '''

        '''
        self.container = container
        self.publicnetwork = publicnetwork
        self.privatenetworks = privatenetworks
        self.forwards = forwards

    def apply_rules(self):
        # nftables
        nftables = templates.render('nftables.conf',
                                    privatenetworks=self.privatenetworks,
                                    publicnetwork=self.publicnetwork,
                                    portforwards=self.forwards)
        self.container.upload_content('/etc/nftables.conf', nftables)
        job = self.container.client.system('nft -f /etc/nftables.conf').get()
        if job.state != 'SUCCESS':
            raise RuntimeError("Failed to apply nftables {} {}".format(job.stdout, job.stderr))



from ..abstracts import AYSable
from js9 import j


class StoragePoolAys(AYSable):

    def __init__(self, storagepool):
        self._obj = storagepool
        self.actor = 'storagepool'

    def create(self, aysrepo):
        try:
            service = aysrepo.serviceGet(role='storagepool', instance=self._obj.name)
        except j.exceptions.NotFound:
            service = None

        device_map, pool_status = self._obj.get_devices_and_status()

        if service is None:
            # create new service
            actor = aysrepo.actorGet(self.actor)

            args = {
                'metadataProfile': self._obj.fsinfo['metadata']['profile'],
                'dataProfile': self._obj.fsinfo['data']['profile'],
                'devices': device_map,
                'node': self._node_name,
                'status': pool_status,
            }
            service = actor.serviceCreate(instance=self._obj.name, args=args)
        else:
            # update model on exists service
            service.model.data.init('devices', len(device_map))
            for i, device in enumerate(device_map):
                service.model.data.devices[i] = device

            service.model.data.status = pool_status
            service.saveAll()

        return service

    @property
    def _node_name(self):
        def is_valid_nic(nic):
            for exclude in ['zt', 'core', 'kvm', 'lo']:
                if nic['name'].startswith(exclude):
                    return False
            return True

        for nic in filter(is_valid_nic, self._obj.node.client.info.nic()):
            if len(nic['addrs']) > 0 and nic['addrs'][0]['addr'] != '':
                return nic['hardwareaddr'].replace(':', '')
        raise AttributeError("name not find for node {}".format(self._obj.node))


class FileSystemAys(AYSable):

    def __init__(self, filesystem):
        self._obj = filesystem
        self.actor = 'filesystem'

    def create(self, aysrepo):
        actor = aysrepo.actorGet(self.actor)
        args = {
            'storagePool': self._obj.pool.name,
            'name': self._obj.name,
            # 'readOnly': ,FIXME
            # 'quota': ,FIXME
        }
        return actor.serviceCreate(instance=self._obj.name, args=args)



from enum import Enum


class EnumJobResultState(Enum):
    unknown_cmd = "unknown_cmd"
    error = "error"
    success = "success"
    killed = "killed"
    timeout = "timeout"
    duplicate_id = "duplicate_id"
    running = "running"



class StorageclustersService:
    def __init__(self, client):
        self.client = client



    def KillCluster(self, label, headers=None, query_params=None, content_type="application/json"):
        """
        Kill cluster
        It is method for DELETE /storageclusters/{label}
        """
        uri = self.client.base_url + "/storageclusters/"+label
        return self.client.delete(uri, None, headers, query_params, content_type)


    def GetClusterInfo(self, label, headers=None, query_params=None, content_type="application/json"):
        """
        Get full information about specific cluster
        It is method for GET /storageclusters/{label}
        """
        uri = self.client.base_url + "/storageclusters/"+label
        return self.client.get(uri, None, headers, query_params, content_type)


    def ListAllClusters(self, headers=None, query_params=None, content_type="application/json"):
        """
        List all running clusters
        It is method for GET /storageclusters
        """
        uri = self.client.base_url + "/storageclusters"
        return self.client.get(uri, None, headers, query_params, content_type)


    def DeployNewCluster(self, data, headers=None, query_params=None, content_type="application/json"):
        """
        Deploy new cluster
        It is method for POST /storageclusters
        """
        uri = self.client.base_url + "/storageclusters"
        return self.client.post(uri, data, headers, query_params, content_type)



from enum import Enum


class EnumJobResultName(Enum):
    core_ping = "core.ping"
    core_system = "core.system"
    core_kill = "core.kill"
    core_killall = "core.killall"
    core_state = "core.state"
    core_reboot = "core.reboot"
    info_cpu = "info.cpu"
    info_disk = "info.disk"
    info_mem = "info.mem"
    info_nic = "info.nic"
    info_os = "info.os"
    container_create = "container.create"
    container_list = "container.list"
    container_dispatch = "container.dispatch"
    container_terminate = "container.terminate"
    bridge_create = "bridge.create"
    bridge_list = "bridge.list"
    bridge_delete = "bridge.delete"
    disk_list = "disk.list"
    disk_mktable = "disk.mktable"
    disk_mkpart = "disk.mkpart"
    disk_rmpart = "disk.rmpart"
    disk_mount = "disk.mount"
    disk_umount = "disk.umount"
    btrfs_create = "btrfs.create"
    btrfs_list = "btrfs.list"
    btrfs_subvol_create = "btrfs.subvol_create"
    btrfs_subvol_list = "btrfs.subvol_list"
    btrfs_subvol_delete = "btrfs.subvol_delete"



"""
Auto-generated class for GetGW
"""
from .GWNIC import GWNIC
from .HTTPProxy import HTTPProxy
from .PortForward import PortForward

from . import client_support


class GetGW(object):
    """
    auto-generated. don't touch.
    """

    @staticmethod
    def create(domain, nics, httpproxies=None, portforwards=None, zerotiernodeid=None):
        """
        :type domain: str
        :type httpproxies: list[HTTPProxy]
        :type nics: list[GWNIC]
        :type portforwards: list[PortForward]
        :type zerotiernodeid: str
        :rtype: GetGW
        """

        return GetGW(
            domain=domain,
            httpproxies=httpproxies,
            nics=nics,
            portforwards=portforwards,
            zerotiernodeid=zerotiernodeid,
        )

    def __init__(self, json=None, **kwargs):
        if json is None and not kwargs:
            raise ValueError('No data or kwargs present')

        class_name = 'GetGW'
        create_error = '{cls}: unable to create {prop} from value: {val}: {err}'
        required_error = '{cls}: missing required property {prop}'

        data = json or kwargs

        property_name = 'domain'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.domain = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'httpproxies'
        val = data.get(property_name)
        if val is not None:
            datatypes = [HTTPProxy]
            try:
                self.httpproxies = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

        property_name = 'nics'
        val = data.get(property_name)
        if val is not None:
            datatypes = [GWNIC]
            try:
                self.nics = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))
        else:
            raise ValueError(required_error.format(cls=class_name, prop=property_name))

        property_name = 'portforwards'
        val = data.get(property_name)
        if val is not None:
            datatypes = [PortForward]
            try:
                self.portforwards = client_support.list_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

        property_name = 'zerotiernodeid'
        val = data.get(property_name)
        if val is not None:
            datatypes = [str]
            try:
                self.zerotiernodeid = client_support.val_factory(val, datatypes)
            except ValueError as err:
                raise ValueError(create_error.format(cls=class_name, prop=property_name, val=val, err=err))

    def __str__(self):
        return self.as_json(indent=4)

    def as_json(self, indent=0):
        return client_support.to_json(self, indent=indent)

    def as_dict(self):
        return client_support.to_dict(self)
